@@ -29,6 +29,7 @@ from __future__ import absolute_import
 import six
 import txaio
 import inspect
+from functools import reduce
 
 from autobahn import wamp
 from autobahn.util import public, IdGenerator, ObservableMixin
@@ -38,7 +39,7 @@ from autobahn.wamp import types
 from autobahn.wamp import role
 from autobahn.wamp import exception
 from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError
-from autobahn.wamp.interfaces import ISession, IPayloadCodec  # noqa
+from autobahn.wamp.interfaces import ISession, IPayloadCodec, IAuthenticator  # noqa
 from autobahn.wamp.types import SessionDetails, CloseDetails, EncodedPayload
 from autobahn.wamp.request import \
     Publication, \
@@ -66,6 +67,8 @@ class BaseSession(ObservableMixin):
     This class implements :class:`autobahn.wamp.interfaces.ISession`.
     """
 
+    log = txaio.make_logger()
+
     def __init__(self):
         """
 
@@ -626,7 +629,7 @@ class ApplicationSession(BaseSession):
                         invoke_kwargs = msg.kwargs if msg.kwargs else dict()
 
                         if handler.details_arg:
-                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo)
+                            invoke_kwargs[handler.details_arg] = types.EventDetails(subscription, msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo)
 
                         # FIXME: https://github.com/crossbario/autobahn-python/issues/764
                         def _success(_):
@@ -887,7 +890,7 @@ class ApplicationSession(BaseSession):
                                 else:
                                     progress = None
 
-                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, caller_authid=msg.caller_authid, caller_authrole=msg.caller_authrole, procedure=proc, enc_algo=msg.enc_algo)
+                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(registration, progress=progress, caller=msg.caller, caller_authid=msg.caller_authid, caller_authrole=msg.caller_authrole, procedure=proc, enc_algo=msg.enc_algo)
 
                             on_reply = txaio.as_future(endpoint.fn, *invoke_args, **invoke_kwargs)
 
@@ -1010,7 +1013,23 @@ class ApplicationSession(BaseSession):
 
             elif isinstance(msg, message.Unregistered):
 
-                if msg.request in self._unregister_reqs:
+                if msg.request == 0:
+                    # this is a forced un-register either from a call
+                    # to the wamp.* meta-api or the force_reregister
+                    # option
+                    try:
+                        reg = self._registrations[msg.registration]
+                    except KeyError:
+                        raise ProtocolError(
+                            "UNREGISTERED received for non-existant registration"
+                            " ID {0}".format(msg.registration)
+                        )
+                    self.log.info(
+                        u"Router unregistered procedure '{proc}' with ID {id}",
+                        proc=reg.procedure,
+                        id=msg.registration,
+                    )
+                elif msg.request in self._unregister_reqs:
 
                     # get and pop outstanding subscribe request
                     request = self._unregister_reqs.pop(msg.request)
@@ -1247,6 +1266,9 @@ class ApplicationSession(BaseSession):
                                       args=args,
                                       kwargs=kwargs)
 
+        if options and options.correlation:
+            msg.correlation = options.correlation
+
         if options and options.acknowledge:
             # only acknowledged publications expect a reply ..
             on_reply = txaio.create_future()
@@ -1294,6 +1316,9 @@ class ApplicationSession(BaseSession):
             else:
                 msg = message.Subscribe(request_id, topic)
 
+            if options and options.correlation:
+                msg.correlation = options.correlation
+
             self._transport.send(msg)
             return on_reply
 
@@ -1310,9 +1335,14 @@ class ApplicationSession(BaseSession):
                 if "_wampuris" in proc.__dict__:
                     for pat in proc.__dict__["_wampuris"]:
                         if pat.is_handler():
-                            uri = pat.uri()
-                            subopts = options or pat.subscribe_options()
-                            on_replies.append(_subscribe(handler, proc, uri, subopts))
+                            _uri = pat.uri()
+                            subopts = pat.options or options
+                            if subopts is None:
+                                if pat.uri_type == uri.Pattern.URI_TYPE_WILDCARD:
+                                    subopts = types.SubscribeOptions(match=u"wildcard")
+                                else:
+                                    subopts = types.SubscribeOptions(match=u"exact")
+                            on_replies.append(_subscribe(handler, proc, _uri, subopts))
 
             # XXX needs coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1404,6 +1434,9 @@ class ApplicationSession(BaseSession):
                                    args=args,
                                    kwargs=kwargs)
 
+        if options and options.correlation:
+            msg.correlation = options.correlation
+
         # FIXME: implement call canceling
         # def canceller(_d):
         #   cancel_msg = message.Cancel(request)
@@ -1431,13 +1464,14 @@ class ApplicationSession(BaseSession):
         return on_reply
 
     @public
-    def register(self, endpoint, procedure=None, options=None):
+    def register(self, endpoint, procedure=None, options=None, prefix=None):
         """
         Implements :func:`autobahn.wamp.interfaces.ICallee.register`
         """
         assert((callable(endpoint) and procedure is not None) or hasattr(endpoint, '__class__'))
         assert(procedure is None or type(procedure) == six.text_type)
         assert(options is None or isinstance(options, types.RegisterOptions))
+        assert prefix is None or isinstance(prefix, six.text_type)
 
         if not self._transport:
             raise exception.TransportLost()
@@ -1446,6 +1480,8 @@ class ApplicationSession(BaseSession):
             request_id = self._request_id_gen.next()
             on_reply = txaio.create_future()
             endpoint_obj = Endpoint(fn, obj, options.details_arg if options else None)
+            if prefix is not None:
+                procedure = u"{}{}".format(prefix, procedure)
             self._register_reqs[request_id] = RegisterRequest(request_id, on_reply, procedure, endpoint_obj)
 
             if options:
@@ -1453,6 +1489,9 @@ class ApplicationSession(BaseSession):
             else:
                 msg = message.Register(request_id, procedure)
 
+            if options and options.correlation:
+                msg.correlation = options.correlation
+
             self._transport.send(msg)
             return on_reply
 
@@ -1470,8 +1509,9 @@ class ApplicationSession(BaseSession):
                 if "_wampuris" in proc.__dict__:
                     for pat in proc.__dict__["_wampuris"]:
                         if pat.is_endpoint():
-                            uri = pat.uri()
-                            on_replies.append(_register(endpoint, proc, uri, options))
+                            _uri = pat.uri()
+                            regopts = pat.options or options
+                            on_replies.append(_register(endpoint, proc, _uri, regopts))
 
             # XXX neds coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1498,6 +1538,131 @@ class ApplicationSession(BaseSession):
         return on_reply
 
 
+# this is NOT public; import from either autobahn.asyncio.wamp or
+# autobahn.twisted.wamp
+class _SessionShim(ApplicationSession):
+    """
+    shim that lets us present pep8 API for user-classes to override,
+    but also backwards-compatible for existing code using
+    ApplicationSession "directly".
+    """
+
+    #: name -> IAuthenticator
+    _authenticators = None
+
+    def onJoin(self, details):
+        return self.on_join(details)
+
+    def onConnect(self):
+        if self._authenticators:
+            # authid, authrole *must* match across all authenticators
+            # (checked in add_authenticator) so these lists are either
+            # [None] or [None, 'some_authid']
+            authid = [x._args.get('authid', None) for x in self._authenticators.values()][-1]
+            authrole = [x._args.get('authrole', None) for x in self._authenticators.values()][-1]
+            authextra = self._merged_authextra()
+            self.join(
+                self.config.realm,
+                authmethods=list(self._authenticators.keys()),
+                authid=authid or u'public',
+                authrole=authrole or u'default',
+                authextra=authextra,
+            )
+        else:
+            self.on_connect()
+
+    def onChallenge(self, challenge):
+        try:
+            authenticator = self._authenticators[challenge.method]
+        except KeyError:
+            raise RuntimeError(
+                "Received challenge for unknown authmethod '{}'".format(
+                    challenge.method
+                )
+            )
+        return authenticator.on_challenge(self, challenge)
+
+    def onLeave(self, details):
+        return self.on_leave(details)
+
+    def onDisconnect(self):
+        return self.on_disconnect()
+
+    # experimental authentication API
+
+    def add_authenticator(self, authenticator):
+        assert isinstance(authenticator, IAuthenticator)
+        if self._authenticators is None:
+            self._authenticators = {}
+
+        # before adding this authenticator we need to validate that
+        # it's consistent with any other authenticators we may have --
+        # for example, they must all agree on "authid" etc because
+        # .join() only takes one value for all of those.
+
+        def at_most_one(name):
+            uni = set([
+                a._args[name]
+                for a in list(self._authenticators.values()) + [authenticator]
+                if name in a._args
+            ])
+            if len(uni) > 1:
+                raise ValueError(
+                    "Inconsistent {}s: {}".format(
+                        name,
+                        ' '.join(uni),
+                    )
+                )
+
+        # all authids must match
+        at_most_one('authid')
+
+        # all authroles must match
+        at_most_one('authrole')
+
+        # can we do anything else other than merge all authextra keys?
+        # here we check that any duplicate keys have the same values
+        authextra = authenticator.authextra
+        merged = self._merged_authextra()
+        for k, v in merged:
+            if k in authextra and authextra[k] != v:
+                raise ValueError(
+                    "Inconsistent authextra values for '{}': '{}' vs '{}'".format(
+                        k, v, authextra[k],
+                    )
+                )
+
+        # validation complete, add it
+        self._authenticators[authenticator.name] = authenticator
+
+    def _merged_authextra(self):
+        authextras = [a._args.get('authextra', {}) for a in self._authenticators.values()]
+        # for all existing _authenticators, we've already checked that
+        # if they contain a key it has the same value as all others.
+        return {
+            k: v
+            for k, v in zip(
+                reduce(lambda x, y: x | set(y.keys()), authextras, set()),
+                reduce(lambda x, y: x | set(y.values()), authextras, set())
+            )
+        }
+
+    # these are the actual "new API" methods (i.e. snake_case)
+    #
+
+    def on_join(self, details):
+        pass
+
+    def on_leave(self, details):
+        self.disconnect()
+
+    def on_connect(self):
+        self.join(self.config.realm)
+
+    def on_disconnect(self):
+        pass
+
+
 # ISession.register collides with the abc.ABCMeta.register method
 # ISession.register(ApplicationSession)
 
